from QuICT.qcda.synthesis.arithmetic.vbe import VBEMulAddMod
from QuICT.qcda.synthesis.arithmetic.bea import BEAMulMod
from QuICT.qcda.synthesis.arithmetic.hrs import HRSMulMod
from QuICT.qcda.synthesis.arithmetic.tmvh import Multiplication
