#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2019/12/13 2:20
# @Author  : Han Yu
# @File    : __version__.py

__title__ = "QuICT"
__version__ = "1.0.0"
__description__ = "quantum algorithm library"
__author__ = "ICT_thoery"
__email__ = "hanyu16@mails.ucas.ac.cn"
__url__ = "http://theory.ict.ac.cn/"
