#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2020/2/11 10:30
# @Author  : Han Yu
# @File    : __init__.py.py

from ._algorithm import Algorithm
from .quantum_algorithm import *
from .quantum_machine_learning import *
from .weight_decision import WeightDecision
